@@ -8,7 +8,10 @@ from python_pachyderm.proto.pfs import pfs_pb2_grpc as pfs_grpc
 from python_pachyderm.proto.pps import pps_pb2 as pps_proto
 from python_pachyderm.proto.pps import pps_pb2_grpc as pps_grpc
 from python_pachyderm.proto.version.versionpb import version_pb2_grpc as version_grpc
-
+from python_pachyderm.proto.transaction import transaction_pb2 as transaction_proto
+from python_pachyderm.proto.transaction import transaction_pb2_grpc as transaction_grpc
+from python_pachyderm.proto.admin import admin_pb2 as admin_proto
+from python_pachyderm.proto.admin import admin_pb2_grpc as admin_grpc
 
 BUFFER_SIZE = 3 * 1024 * 1024  # 3MB TODO: Base this on some grpc value
 
@@ -27,6 +30,13 @@ def commit_from(src, allow_just_repo=False):
     return pfs_proto.Commit(repo=pfs_proto.Repo(name=src))
 
 
+def transaction_from(transaction):
+    if isinstance(transaction, transaction_proto.Transaction):
+        return transaction
+    else:
+        return transaction_proto.Transaction(id=transaction)
+
+
 class Client(object):
     def __init__(self, host=None, port=None, auth_token=None, root_certs=None):
         """
@@ -38,7 +48,7 @@ class Client(object):
         `pachctl port-forward`.
         * `port`: The port to connect to. Default is 30650.
         * `auth_token`: The authentication token; used if authentication is
-        enabled on the cluster. Default to `None`.
+        enabled on the cluster. Defaults to `None`.
         * `root_certs`:  The PEM-encoded root certificates as byte string.
         """
 
@@ -56,42 +66,45 @@ class Client(object):
 
         self.root_certs = root_certs
 
+    def _create_stub(self, grpc_module):
+        if self.root_certs:
+            ssl_channel_credentials = grpc_module.grpc.ssl_channel_credentials
+            ssl = ssl_channel_credentials(root_certificates=self.root_certs)
+            channel = grpc_module.grpc.secure_channel(self.address, ssl)
+        else:
+            channel = grpc_module.grpc.insecure_channel(self.address)
+        return grpc_module.APIStub(channel)
+
     @property
     def _pfs_stub(self):
         if not hasattr(self, "__pfs_stub"):
-            if self.root_certs:
-                ssl_channel_credentials = pfs_grpc.grpc.ssl_channel_credentials
-                ssl = ssl_channel_credentials(root_certificates=self.root_certs)
-                channel = pfs_grpc.grpc.secure_channel(self.address, ssl)
-            else:
-                channel = pfs_grpc.grpc.insecure_channel(self.address)
-            self.__pfs_stub = pfs_grpc.APIStub(channel)
+            self.__pfs_stub = self._create_stub(pfs_grpc)
         return self.__pfs_stub
 
     @property
     def _pps_stub(self):
         if not hasattr(self, "__pps_stub"):
-            if self.root_certs:
-                ssl_channel_credentials = pps_grpc.grpc.ssl_channel_credentials
-                ssl = ssl_channel_credentials(root_certificates=self.root_certs)
-                channel = pps_grpc.grpc.secure_channel(self.address, ssl)
-            else:
-                channel = pps_grpc.grpc.insecure_channel(self.address)
-            self.__pps_stub = pps_grpc.APIStub(channel)
+            self.__pps_stub = self._create_stub(pps_grpc)
         return self.__pps_stub
 
     @property
     def _version_stub(self):
         if not hasattr(self, "__version_stub"):
-            if self.root_certs:
-                ssl_channel_credentials = version_grpc.grpc.ssl_channel_credentials
-                ssl = ssl_channel_credentials(root_certificates=self.root_certs)
-                channel = version_grpc.grpc.secure_channel(self.address, ssl)
-            else:
-                channel = version_grpc.grpc.insecure_channel(self.address)
-            self.__version_stub = version_grpc.APIStub(channel)
+            self.__version_stub = self._create_stub(version_grpc)
         return self.__version_stub
 
+    @property
+    def _transaction_stub(self):
+        if not hasattr(self, "__transaction_stub"):
+            self.__transaction_stub = self._create_stub(transaction_grpc)
+        return self.__transaction_stub
+
+    @property
+    def _admin_stub(self):
+        if not hasattr(self, "__admin_stub"):
+            self.__admin_stub = self._create_stub(admin_grpc)
+        return self.__admin_stub
+
     def get_remote_version(self):
         req = version_grpc.google_dot_protobuf_dot_empty__pb2.Empty()
         return self._version_stub.GetVersion(req, metadata=self.metadata)
@@ -789,7 +802,7 @@ class Client(object):
                         enable_stats=None, reprocess=None, max_queue_size=None,
                         service=None, chunk_spec=None, datum_timeout=None,
                         job_timeout=None, salt=None, standby=None, datum_tries=None,
-                        scheduling_spec=None, pod_patch=None):
+                        scheduling_spec=None, pod_patch=None, spout=None):
         """
         Creates a pipeline. For more info, please refer to the pipeline spec
         document:
@@ -825,6 +838,7 @@ class Client(object):
         * `datum_tries`: An optional int.
         * `scheduling_spec`: An optional `SchedulingSpec` object.
         * `pod_patch`: An optional string.
+        * `spout`: An optional `Spout` object.
         """
 
         req = pps_proto.CreatePipelineRequest(
@@ -839,7 +853,7 @@ class Client(object):
             chunk_spec=chunk_spec, datum_timeout=datum_timeout,
             job_timeout=job_timeout, salt=salt, standby=standby,
             datum_tries=datum_tries, scheduling_spec=scheduling_spec,
-            pod_patch=pod_patch
+            pod_patch=pod_patch, spout=spout,
         )
         self._pps_stub.CreatePipeline(req, metadata=self.metadata)
 
@@ -1023,3 +1037,134 @@ class Client(object):
         Runs garbage collection.
         """
         return self._pps_stub.GarbageCollect(pps_proto.GarbageCollectRequest(), metadata=self.metadata)
+
+    def start_transaction(self):
+        """
+        Starts a transaction.
+        """
+        req = transaction_proto.StartTransactionRequest()
+        return self._transaction_stub.StartTransaction(req, metadata=self.metadata)
+
+    def inspect_transaction(self, transaction):
+        """
+        Inspects a given transaction.
+
+        Params:
+
+        * `transaction`: A string or `Transaction` object.
+        """
+        req = transaction_proto.InspectTransactionRequest(transaction=transaction_from(transaction))
+        return self._transaction_stub.InspectTransaction(req, metadata=self.metadata)
+
+    def delete_transaction(self, transaction):
+        """
+        Deletes a given transaction.
+
+        Params:
+
+        * `transaction`: A string or `Transaction` object.
+        """
+        req = transaction_proto.DeleteTransactionRequest(transaction=transaction_from(transaction))
+        return self._transaction_stub.DeleteTransaction(req, metadata=self.metadata)
+
+    def list_transaction(self):
+        """
+        Lists transactions.
+        """
+        req = transaction_proto.ListTransactionRequest()
+        return self._transaction_stub.ListTransaction(req, metadata=self.metadata).transaction_info
+
+    def finish_transaction(self, transaction):
+        """
+        Finishes a given transaction.
+
+        Params:
+
+        * `transaction`: A string or `Transaction` object.
+        """
+        req = transaction_proto.FinishTransactionRequest(transaction=transaction_from(transaction))
+        return self._transaction_stub.FinishTransaction(req, metadata=self.metadata)
+
+    @contextmanager
+    def transaction(self, transaction=None):
+        """
+        A context manager for running operations within a transaction. When
+        the context manager completes, the transaction will be deleted if an
+        error occurred, or otherwise finished.
+
+        Params:
+
+        * `transaction`: An optional string or `Transaction` object. If
+        unspecified, a new transaction will be started.
+        """
+
+        # note that this is different from `pachctl`, which will delete any
+        # active transaction
+        for (k, v) in self.metadata:
+            if k == "pach-transaction":
+                raise Exception("this client already has an active transaction with ID={}".format(v))
+
+        if transaction is None:
+            transaction = self.start_transaction()
+        else:
+            transaction = transaction_from(transaction)
+
+        self.metadata.append(("pach-transaction", transaction.id))
+
+        try:
+            yield transaction
+        except Exception:
+            self.delete_transaction(transaction)
+            raise
+        else:
+            self.finish_transaction(transaction)
+        finally:
+            self.metadata = [(k, v) for (k, v) in self.metadata if k != "pach-transaction"]
+
+    def extract(self, url=None, no_objects=None, no_repos=None, no_pipelines=None):
+        """
+        Extracts cluster data for backup. Yields `Op` objects.
+
+        Params:
+
+        * `url`: An optional string specifying an object storage URL. If set,
+          data will be extracted to this URL rather than returned.
+        * `no_objects`: An optional bool. If true, will cause extract to omit
+          objects (and tags.)
+        * `no_repos`: An optional bool. If true, will cause extract to omit
+          repos, commits and branches.
+        * `no_pipelines`: An optional bool. If true, will cause extract to
+          omit pipelines.
+        """
+        url = url or ""
+        req = admin_proto.ExtractRequest(URL=url, no_objects=no_objects, no_repos=no_repos, no_pipelines=no_pipelines)
+        return self._admin_stub.Extract(req, metadata=self.metadata)
+
+    def extract_pipeline(self, pipeline_name):
+        """
+        Extracts a pipeline for backup. Returns an `Op` object.
+
+        Params:
+
+        * `pipeline_name`: A string representing a pipeline name to extract.
+        """
+        pipeline = pps_proto.Pipeline(name=pipeline_name) if pipeline_name is not None else None
+        req = admin_proto.ExtractPipelineRequest(pipeline=pipeline)
+        return self._admin_stub.ExtractPipeline(req, metadata=self.metadata)
+
+    def restore(self, requests):
+        """
+        Restores a cluster.
+
+        Params:
+
+        * `requests`: A generator of `RestoreRequest` objects.
+        """
+        self._admin_stub.Restore(requests, metadata=self.metadata)
+
+    def inspect_cluster(self):
+        """
+        Inspects a cluster. Returns a `ClusterInfo` object.
+        """
+        req = version_grpc.google_dot_protobuf_dot_empty__pb2.Empty()
+        return self._admin_stub.InspectCluster(req, metadata=self.metadata)
